#include <ch32v00x.h>
#include <stdbool.h>
#include <debug.h>

#define POSITION_NEUTRAL 1500
#define POSITION_MINIMUM 500
#define POSITION_MAXIMUM 2500
#define POSITION_INCREMENT 100

volatile bool encoderPulse;
volatile int dirFlag;

__attribute__((interrupt("WCH-Interrupt-fast")))
void TIM2_IRQHandler() {
	if (TIM_GetITStatus(TIM2, TIM_IT_Update) != RESET) {
		dirFlag = (TIM2->CTLR1 & (1 << 4)) >> 4;
		encoderPulse = true;
		TIM_ClearITPendingBit(TIM2, TIM_IT_Update);
	}
}

int main() {
	USART_Printf_Init(115200);

	// -- On configure les entrées sur lesquelles est connecté l'encodeur.
	// PD4 = T2CH1 = CLK
	// PD3 = T2CH2 = DT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOD, ENABLE);
	GPIO_InitTypeDef gpioInit = { 0 };
	gpioInit.GPIO_Pin = GPIO_Pin_4 | GPIO_Pin_3;
	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &gpioInit);
	// et la sortie PWM (PD2 = T1CH1).
	gpioInit.GPIO_Pin = GPIO_Pin_2;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOD, &gpioInit);

	// -- On configure TIM2 en mode encodeur.
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit = { 0 };
	timeBaseInit.TIM_Period = 1;
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV4;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM2, &timeBaseInit);

	TIM_EncoderInterfaceConfig(TIM2, TIM_EncoderMode_TI2, TIM_ICPolarity_Falling, TIM_ICPolarity_Falling);

	TIM_ICInitTypeDef icInit;
	icInit.TIM_Channel = TIM_Channel_1;
	icInit.TIM_ICPolarity = TIM_ICPolarity_Falling;
	icInit.TIM_ICSelection = TIM_ICSelection_DirectTI;
	icInit.TIM_ICPrescaler = TIM_ICPSC_DIV1;
	icInit.TIM_ICFilter = 12;
	TIM_ICInit(TIM2, &icInit);

	NVIC_InitTypeDef nvicInit = { 0 };
	nvicInit.NVIC_IRQChannel = TIM2_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = 0;
	nvicInit.NVIC_IRQChannelSubPriority = 1;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);
	TIM_ITConfig(TIM2, TIM_IT_Update, ENABLE);

	TIM_Cmd(TIM2, ENABLE);

	// On configure TIM1 en mode PWM.
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);

	// On veut exprimer les durées en microsecondes,
	// donc on doit diviser l'horloge principale par 48.
	timeBaseInit.TIM_Prescaler = 48 - 1;
	// On veut une fréquence du signal de sortie de 50Hz,
	// soit une période de 20000 microsecondes.
	timeBaseInit.TIM_Period = 20000 - 1;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM1, &timeBaseInit);

	TIM_OCInitTypeDef ocInit = { 0 };
	ocInit.TIM_OCMode = TIM_OCMode_PWM1;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_Pulse = POSITION_NEUTRAL;
	TIM_OC1Init(TIM1, &ocInit);

	TIM_CtrlPWMOutputs(TIM1, ENABLE);
	TIM_Cmd(TIM1, ENABLE);

	int servoPosition = POSITION_NEUTRAL;
	encoderPulse = false;
	printf("position = %d microsecondes\r\n", servoPosition);

	while (1) {
		if (encoderPulse) {
			encoderPulse = false;

			if (dirFlag) {
				if (servoPosition <= (POSITION_MAXIMUM - POSITION_INCREMENT)) {
					servoPosition += POSITION_INCREMENT;
				}
			} else {
				if (servoPosition >= (POSITION_MINIMUM + POSITION_INCREMENT)) {
					servoPosition -= POSITION_INCREMENT;
				}
			}

			TIM_SetCompare1(TIM1, servoPosition);
			printf("position = %d microsecondes\r\n", servoPosition);
		}
	}
}
